// This file is part of river-bedload
//
// Copyright (C) 2023 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const assert = std.debug.assert;
const builtin = @import("builtin");
const fmt = std.fmt;
const io = std.io;
const json = std.json;
const mem = std.mem;
const posix = std.posix;

const wayland = @import("wayland");
const wl = wayland.client.wl;
const zriver = wayland.client.zriver;

const ctx = &@import("main.zig").context;
const Backend = @This();

const log = std.log.scoped(.backend);

const Tag = struct {
    output: ?[]const u8 = null,
    id: u32 = undefined,

    active: bool = false,
    focused: bool = false,
    occupied: bool = false,
    urgent: bool = false,
};

const OutputInfo = struct {
    name: ?[]const u8 = null,
    description: ?[]const u8 = null,
    focused: bool = false,

    layout_name: ?[]const u8 = null,

    focused_tags: u32 = 0,
    view_tags: u32 = 0,
    urgent_tags: u32 = 0,
};

const Output = struct {
    wl_output: *wl.Output = undefined,
    wl_name: u32 = undefined,
    river_output_status: *zriver.OutputStatusV1 = undefined,
    configured: bool = false,

    tags: std.ArrayListUnmanaged(Tag) = .{},

    info: OutputInfo = .{},

    fn init(output: *Output, wl_output: *wl.Output, wl_name: u32) !void {
        output.* = .{
            .wl_output = wl_output,
            .wl_name = wl_name,
        };

        try output.tags.appendNTimes(ctx.gpa, Tag{}, 32);
        output.wl_output.setListener(*Output, output_listener, output);
    }

    fn destroy(output: *Output) void {
        output.wl_output.release();
        output.river_output_status.destroy();
        output.tags.deinit(ctx.gpa);
        if (output.info.name) |name| ctx.gpa.free(name);
        if (output.info.description) |description| ctx.gpa.free(description);
        if (output.info.layout_name) |name| ctx.gpa.free(name);

        const node: *std.SinglyLinkedList(Output).Node = @fieldParentPtr("data", output);
        ctx.backend.outputs.remove(node);
        ctx.gpa.destroy(node);
    }

    fn get_output_status(output: *Output) !void {
        output.river_output_status =
            try ctx.backend.river_status_manager.?.getRiverOutputStatus(output.wl_output);
        output.river_output_status.setListener(*Output, output_status_listener, output);
        output.configured = true;
    }

    fn output_listener(_: *wl.Output, event: wl.Output.Event, output: *Output) void {
        switch (event) {
            .name => |ev| {
                output.info.name = ctx.gpa.dupe(u8, mem.span(ev.name)) catch @panic("out of memory");
            },
            .description => |ev| {
                output.info.description = ctx.gpa.dupe(u8, mem.span(ev.description)) catch @panic("out of memory");
            },
            else => {},
        }
    }

    fn update_tags(output: *Output) !void {
        var i: u32 = 0;
        while (i < 32) : (i += 1) {
            const focused = (output.info.focused_tags & (@as(u32, 1) << @as(u5, @intCast(i)))) != 0;
            const urgent = (output.info.urgent_tags & (@as(u32, 1) << @as(u5, @intCast(i)))) != 0;
            const occupied = (output.info.view_tags & (@as(u32, 1) << @as(u5, @intCast(i)))) != 0;
            const active = focused or urgent or occupied;

            output.tags.items[i].output = output.info.name;
            output.tags.items[i].id = i + 1;
            output.tags.items[i].active = active;
            output.tags.items[i].focused = focused;
            output.tags.items[i].urgent = urgent;
            output.tags.items[i].occupied = occupied;
        }
    }

    fn handle_focused_tags(output: *Output, tags: u32) !void {
        output.info.focused_tags = tags;
        try output.update_tags();
    }

    fn handle_view_tags(output: *Output, tags: *wl.Array) !void {
        output.info.view_tags = 0;
        for (tags.slice(u32)) |tag| {
            output.info.view_tags |= tag;
        }
        try output.update_tags();
    }

    fn handle_urgent_tags(output: *Output, tags: u32) !void {
        output.info.urgent_tags = tags;
        try output.update_tags();
    }

    fn output_status_listener(
        _: *zriver.OutputStatusV1,
        event: zriver.OutputStatusV1.Event,
        output: *Output,
    ) void {
        switch (event) {
            .focused_tags => |ev| output.handle_focused_tags(ev.tags) catch return,
            .view_tags => |ev| output.handle_view_tags(ev.tags) catch return,
            .urgent_tags => |ev| output.handle_urgent_tags(ev.tags) catch return,
            .layout_name => |ev| {
                output.info.layout_name = ctx.gpa.dupe(u8, mem.span(ev.name)) catch @panic("out of memory");
            },
            .layout_name_clear => if (output.info.layout_name) |name| ctx.gpa.free(name),
        }
    }
};

const SeatInfo = struct {
    name: ?[]const u8 = null,
    focused_view: ?[]const u8 = null,
    mode: []const u8 = undefined,
};

const Seat = struct {
    wl_seat: *wl.Seat,
    id: u32,
    river_seat_status: *zriver.SeatStatusV1 = undefined,
    configured: bool = false,

    info: SeatInfo = .{},

    fn init(seat: *Seat, wl_seat: *wl.Seat, id: u32) !void {
        seat.* = .{
            .wl_seat = wl_seat,
            .id = id,
        };

        seat.wl_seat.setListener(*Seat, seat_listener, seat);
    }

    fn destroy(seat: *Seat) void {
        seat.river_seat_status.destroy();
        seat.wl_seat.release();

        if (seat.info.name) |name| ctx.gpa.free(name);
        if (seat.info.focused_view) |fs| ctx.gpa.free(fs);
        ctx.gpa.free(seat.info.mode);

        const node: *std.SinglyLinkedList(Seat).Node = @fieldParentPtr("data", seat);
        ctx.backend.seats.remove(node);
        ctx.gpa.destroy(node);
    }

    fn get_seat_status(seat: *Seat) !void {
        seat.river_seat_status =
            try ctx.backend.river_status_manager.?.getRiverSeatStatus(seat.wl_seat);
        seat.river_seat_status.setListener(*Seat, seat_status_listener, seat);
        seat.configured = true;
    }

    fn seat_listener(_: *wl.Seat, event: wl.Seat.Event, seat: *Seat) void {
        switch (event) {
            .name => |ev| {
                seat.info.name = ctx.gpa.dupe(u8, mem.span(ev.name)) catch @panic("out of memory");
            },
            else => {},
        }
    }

    fn seat_status_listener(_: *zriver.SeatStatusV1, event: zriver.SeatStatusV1.Event, seat: *Seat) void {
        switch (event) {
            .focused_output => |ev| {
                var it = ctx.backend.outputs.first;
                while (it) |node| : (it = node.next) {
                    if (ev.output == node.data.wl_output) {
                        node.data.info.focused = true;
                        break;
                    }
                }
            },
            .unfocused_output => |ev| {
                var it = ctx.backend.outputs.first;
                while (it) |node| : (it = node.next) {
                    if (ev.output == node.data.wl_output) {
                        node.data.info.focused = false;
                        break;
                    }
                }
            },
            .focused_view => |ev| {
                seat.info.focused_view = ctx.gpa.dupe(u8, mem.span(ev.title)) catch @panic("out of memory");
            },
            .mode => |ev| {
                seat.info.mode = ctx.gpa.dupe(u8, mem.span(ev.name)) catch @panic("out of memory");
            },
        }
    }
};

pub const Sort = enum {
    outputs,
    seats,

    tags,
    active,
    focused,
    urgent,

    title,
    layout,
};

display: *wl.Display = undefined,
registry: *wl.Registry = undefined,

compositor: ?*wl.Compositor = null,
shm: ?*wl.Shm = null,

river_status_manager: ?*zriver.StatusManagerV1 = null,

outputs: std.SinglyLinkedList(Output) = .{},
seats: std.SinglyLinkedList(Seat) = .{},

sort: Sort = undefined,
minified: bool = false,

pub fn init(backend: *Backend) !bool {
    backend.* = .{
        .display = wl.Display.connect(null) catch {
            ctx.fatal(.backend, "failed to connect to Wayland compositor", .{});
        },
        .registry = try backend.display.getRegistry(),
    };

    backend.registry.setListener(*Backend, registry_listener, backend);

    const errno = backend.display.roundtrip();
    if (errno != .SUCCESS) {
        ctx.fatal(.backend, "initial roundtrip failed: E{s}", .{@tagName(errno)});
    }

    if (backend.compositor == null) fatal_not_advertised(wl.Compositor);
    if (backend.shm == null) fatal_not_advertised(wl.Shm);

    // TODO: That should not be here but since river_status_manager is bound
    // after wl_seat, it's not called in registry_event()
    var it = backend.seats.first;
    while (it) |node| : (it = node.next) {
        if (backend.river_status_manager != null) try node.data.get_seat_status();
    }

    return backend.display.roundtrip() == .SUCCESS;
}

pub fn deinit(backend: *Backend) void {
    if (backend.compositor) |cmp| cmp.destroy();
    if (backend.shm) |shm| shm.destroy();
    if (backend.river_status_manager) |rsm| rsm.destroy();

    while (backend.outputs.first) |node| node.data.destroy();
    while (backend.seats.first) |node| node.data.destroy();

    backend.registry.destroy();
    backend.display.disconnect();

    backend.* = undefined;
}

pub fn print_json(backend: *Backend, out: anytype) !void {
    const opts = json.StringifyOptions{
        .whitespace = if (backend.minified) .minified else .indent_tab,
    };

    var arena = std.heap.ArenaAllocator.init(ctx.gpa);
    defer arena.deinit();
    const ally = arena.allocator();

    switch (backend.sort) {
        .outputs => {
            const list = try ally.alloc(OutputInfo, ctx.backend.outputs.len());

            var it = backend.outputs.first;
            var i: usize = 0;
            while (it) |node| : (it = node.next) {
                list[i] = node.data.info;
                i += 1;
            }

            try json.stringify(list, opts, out);
        },
        .seats => {
            const list = try ally.alloc(SeatInfo, ctx.backend.seats.len());

            var it = backend.seats.first;
            var i: usize = 0;
            while (it) |node| : (it = node.next) {
                list[i] = node.data.info;
                i += 1;
            }

            try json.stringify(list, opts, out);
        },
        .tags => {
            var list = try std.ArrayList(Tag).initCapacity(ally, 32 * backend.outputs.len());

            var it = backend.outputs.first;
            while (it) |node| : (it = node.next) {
                try list.appendSlice(node.data.tags.items);
            }

            try json.stringify(list.items, opts, out);
        },
        .active => {
            var list: std.ArrayListUnmanaged(Tag) = .{};

            var it = backend.outputs.first;
            while (it) |node| : (it = node.next) {
                for (node.data.tags.items) |tag| {
                    if (tag.active) try list.append(ally, tag);
                }
            }

            try json.stringify(list.items, opts, out);
        },
        .focused => {
            const Slice = struct {
                output: ?[]const u8 = null,
                focused_id: u32,
            };
            const list = try ally.alloc(Slice, backend.outputs.len());

            var it = backend.outputs.first;
            var i: usize = 0;
            while (it) |node| : (it = node.next) {
                for (node.data.tags.items) |tag| {
                    if (tag.focused) {
                        list[i] = .{ .output = node.data.info.name, .focused_id = tag.id };
                        break;
                    }
                }
                i += 1;
            }

            try json.stringify(list, opts, out);
        },
        .urgent => {
            const Slice = struct {
                output: ?[]const u8 = null,
                urgent_id: u32 = undefined,
            };
            var list: std.ArrayListUnmanaged(Slice) = .{};

            var it = backend.outputs.first;
            while (it) |node| : (it = node.next) {
                for (node.data.tags.items) |tag| {
                    if (tag.urgent) {
                        try list.append(ally, .{ .output = node.data.info.name, .urgent_id = tag.id });
                    }
                }
            }

            try json.stringify(list.items, opts, out);
        },
        .title => {
            const Slice = struct {
                title: ?[]const u8 = null,
            };
            const list = try ally.alloc(Slice, ctx.backend.seats.len());

            var it = backend.seats.first;
            var i: usize = 0;
            while (it) |node| : (it = node.next) {
                list[i] = .{ .title = node.data.info.focused_view };
                i += 1;
            }

            try json.stringify(list, opts, out);
        },
        .layout => {
            const Slice = struct {
                output: ?[]const u8 = null,
                layout: ?[]const u8 = null,
            };
            const list = try ally.alloc(Slice, ctx.backend.outputs.len());

            var it = backend.outputs.first;
            var i: usize = 0;
            while (it) |node| : (it = node.next) {
                list[i] = .{ .output = node.data.info.name, .layout = node.data.info.layout_name };
                i += 1;
            }

            try json.stringify(list, opts, out);
        },
    }
    try out.writeByte('\n');
}

/// REF: https://github.com/ifreund/waylock/blob/master/src/Lock.zig#L184
/// This function does the following:
///  1. Dispatch buffered wayland events to their listener callbacks.
///  2. Prepare the wayland connection for reading.
///  3. Send all buffered wayland requests to the server.
/// After this function has been called, either wl.Display.readEvents() or
/// wl.Display.cancelRead() read must be called.
pub fn flush_wayland_and_prepare_read(backend: *Backend) void {
    while (!backend.display.prepareRead()) {
        const errno = backend.display.dispatchPending();
        if (errno != .SUCCESS) {
            ctx.fatal(.backend, "failed to dispatch pending wayland events: E{s}", .{@tagName(errno)});
        }
    }

    while (true) {
        const errno = backend.display.flush();
        switch (errno) {
            .SUCCESS => return,
            .PIPE => {
                // libwayland uses this error to indicate that the wayland server
                // closed its side of the wayland socket. We want to continue to
                // read any buffered messages from the server though as there is
                // likely a protocol error message we'd like libwayland to log.
                _ = backend.display.readEvents();
                ctx.fatal(.backend, "connection to wayland server unexpectedly terminated", .{});
            },
            .AGAIN => {
                // The socket buffer is full, so wait for it to become writable again.
                var wayland_out = [_]posix.pollfd{.{
                    .fd = backend.display.getFd(),
                    .events = posix.POLL.OUT,
                    .revents = 0,
                }};
                _ = posix.poll(&wayland_out, -1) catch |err| {
                    ctx.fatal(.backend, "poll() failed: {s}", .{@errorName(err)});
                };
                // No need to check for POLLHUP/POLLERR here, just fall
                // through to the next flush() to handle them in one place.
            },
            else => {
                ctx.fatal(.backend, "failed to flush wayland requests: E{s}", .{@tagName(errno)});
            },
        }
    }
}

fn registry_listener(registry: *wl.Registry, event: wl.Registry.Event, backend: *Backend) void {
    backend.registry_event(registry, event) catch |err| switch (err) {
        error.OutOfMemory => {
            log.err("out of memory", .{});
            return;
        },
        else => return,
    };
}

fn registry_event(backend: *Backend, registry: *wl.Registry, event: wl.Registry.Event) !void {
    switch (event) {
        .global => |ev| {
            if (mem.orderZ(u8, ev.interface, wl.Compositor.getInterface().name) == .eq) {
                backend.compositor = try registry.bind(ev.name, wl.Compositor, 5);
            } else if (mem.orderZ(u8, ev.interface, wl.Shm.getInterface().name) == .eq) {
                backend.shm = try registry.bind(ev.name, wl.Shm, 1);
            } else if (mem.orderZ(u8, ev.interface, wl.Output.getInterface().name) == .eq) {
                // Version 4 required for wl_output.name.
                if (ev.version < 4) fatal_version(wl.Output, ev.version, 4);

                const wl_output = try registry.bind(ev.name, wl.Output, 4);
                errdefer wl_output.release();

                const node = try ctx.gpa.create(std.SinglyLinkedList(Output).Node);
                errdefer ctx.gpa.destroy(node);

                try node.data.init(wl_output, ev.name);
                backend.outputs.prepend(node);

                if (backend.river_status_manager != null) try node.data.get_output_status();
            } else if (mem.orderZ(u8, ev.interface, wl.Seat.getInterface().name) == .eq) {
                // Version 5 required for wl_seat.release.
                if (ev.version < 5) fatal_version(wl.Seat, ev.version, 5);

                const wl_seat = try registry.bind(ev.name, wl.Seat, 8);
                errdefer wl_seat.release();

                const node = try ctx.gpa.create(std.SinglyLinkedList(Seat).Node);
                errdefer ctx.gpa.destroy(node);

                try node.data.init(wl_seat, ev.name);
                backend.seats.prepend(node);

                if (backend.river_status_manager != null) try node.data.get_seat_status();
            } else if (mem.orderZ(u8, ev.interface, zriver.StatusManagerV1.getInterface().name) == .eq) {
                backend.river_status_manager = try registry.bind(ev.name, zriver.StatusManagerV1, 4);
            }
        },
        .global_remove => |ev| {
            {
                var it = backend.outputs.first;
                while (it) |node| : (it = node.next) {
                    if (node.data.wl_name == ev.name) {
                        node.data.destroy();
                        break;
                    }
                }
            }

            {
                var it = backend.seats.first;
                while (it) |node| : (it = node.next) {
                    if (node.data.id == ev.name) {
                        node.data.destroy();
                        break;
                    }
                }
            }
        },
    }
}

fn fatal_version(comptime Global: type, found_version: u32, req_version: u32) noreturn {
    ctx.fatal(
        .backend,
        "the advertised {s} version is too old({d}). Version {d} is required",
        .{ Global.getInterface().name, found_version, req_version },
    );
}

fn fatal_not_advertised(comptime Global: type) noreturn {
    ctx.fatal(.backend, "{s} not advertised", .{Global.getInterface().name});
}
