// This file is part of river-bedload
//
// Copyright (C) 2023 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const mem = std.mem;
const posix = std.posix;

const Backend = @import("Backend.zig");
const event_loop = @import("event_loop.zig");
const flags = @import("flags");

const usage =
    \\Usage: river-bedload [options...]
    \\
    \\  -h              Print this help message and exit.
    \\  -watch <sort>   Print to STDOUT in json format.
    \\  -print <sort>   Print to STDOUT in json format and exit.
    \\  -minfied        Print with no whitespace.
    \\
    \\sort:
    \\  outputs
    \\  seats
    \\  tags: All tags from from all outputs
    \\  active: Only active tags (i.e. focused occupied or urgent)
    \\  focused: ID of the focused tags
    \\  urgent: ID of the focused tags
    \\  title: Focused windows title
    \\  layout: Layout name of outputs
    \\
;

const Context = struct {
    gpa: mem.Allocator = std.heap.c_allocator,
    backend: Backend = .{},

    initialized: bool = false,

    fn deinit(ctx: *Context) void {
        ctx.backend.deinit();
    }

    pub fn fatal(
        _: Context,
        comptime scope: @Type(.EnumLiteral),
        comptime format: []const u8,
        args: anytype,
    ) noreturn {
        std.log.scoped(scope).err(format, args);
        posix.exit(1);
    }
};

pub var context: Context = .{};

pub fn main() anyerror!void {
    const res = flags.parser([*:0]const u8, &.{
        .{ .name = "h", .kind = .boolean },
        .{ .name = "minified", .kind = .boolean },
        .{ .name = "watch", .kind = .arg },
        .{ .name = "print", .kind = .arg },
    }).parse(std.os.argv[1..]) catch {
        try std.io.getStdErr().writeAll(usage);
        posix.exit(1);
    };
    if (res.args.len != 0) fatal_usage("Unknown option '{s}'", .{res.args[0]});

    if (res.flags.h) {
        try std.io.getStdOut().writeAll(usage);
        posix.exit(0);
    }

    context.initialized = try context.backend.init();
    defer context.deinit();

    std.debug.assert(context.initialized);

    if (res.flags.watch != null and res.flags.print != null) {
        fatal_usage("Only -watch OR -print allowed'", .{});
    }
    if (res.flags.minified) context.backend.minified = true;
    if (res.flags.watch) |raw| {
        if (std.meta.stringToEnum(Backend.Sort, raw)) |sort| {
            context.backend.sort = sort;
        } else {
            fatal_usage("Unknown -watch option '{s}'", .{raw});
        }

        try event_loop.run();
    }
    if (res.flags.print) |raw| {
        if (std.meta.stringToEnum(Backend.Sort, raw)) |sort| {
            context.backend.sort = sort;
            var bw = std.io.bufferedWriter(std.io.getStdOut().writer());
            const w = bw.writer();
            try context.backend.print_json(w);
            try bw.flush();
        } else {
            fatal_usage("Unknown -watch option '{s}'", .{raw});
        }
    }
}

fn fatal_usage(comptime format: []const u8, args: anytype) noreturn {
    std.log.err(format, args);
    std.io.getStdErr().writeAll(usage) catch {};
    posix.exit(1);
}
